#include "IOTCDevice.h"
#include "RDTServer.h"
#include "common.h"
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#define RDT_USERNAME "admin"
#define RDT_PASSWORD "testpwd"

int gFd;
static char gFn[64];

void RDT_Server_Authentication(const char *cszAccount, char *cszPassword, uint32_t nPasswordMaxLength, const void *pArg) {
    printf("[%s] cszAccount=%s, nPasswordMaxLength=%d, pArg=%s\n", __func__, cszAccount, nPasswordMaxLength, (char *)pArg);
    if (strcmp(cszAccount, RDT_USERNAME) == 0) {
        strcpy(cszPassword, RDT_PASSWORD);
    } else {
        printf("unknow account, reject connect\n");
        // do nothing to reject connect
    }
}

void *thread_Send(void *arg) {
    FILE *fp = fopen(gFn, "rb");
    if (fp == NULL) {
        printf("FOpen file[%s] errno[%d]!\n", gFn, errno);
        pthread_exit(0);
    }

    int SID = *(int *)arg;
    free(arg);
    struct st_SInfoEx sInfo;
    sInfo.size = sizeof(struct st_SInfoEx);
    IOTC_Session_Check_Ex(SID, &sInfo);
    char buf[MAX_BUF_SIZE];

    char *args = strdup("test_args");
    int RDT_ID = RDT_Server_Create(SID, 0, RDT_Server_Authentication, args, 3000);
    if (RDT_ID < 0) {
        printf("RDT_Create failed[%d]!!\n", RDT_ID);
        IOTC_Session_Close(SID);
        fclose(fp);
        pthread_exit(0);
    }

    printf("RDT_ID = %d\n", RDT_ID);

    int ret = RDT_Read(RDT_ID, buf, 1024, RDT_WAIT_TIMEMS);
    if (ret < 0) {
        printf("RDT rcv cmd error[%d]!\n", ret);
        fclose(fp);
        pthread_exit(0);
    }
    int requestCnt = atoi(buf), i;
    struct stat fStat;
    fstat(gFd, &fStat);
    struct st_RDT_Status rdtStatus;
    printf("requestCnt[%d] file size[%lu]\n", requestCnt, fStat.st_size);
    sprintf(buf, "%lu", fStat.st_size);
    ret = RDT_Write(RDT_ID, buf, strlen(buf));
    if (ret < 0) {
        printf("RDT send file size error[%d]!\n", ret);
        fclose(fp);
        pthread_exit(0);
    }
    ret = RDT_Read(RDT_ID, buf, 1024, RDT_WAIT_TIMEMS);
    if (ret < 0 && strcmp(buf, START_STRING) != 0) {
        printf("RDT rcv start to send file error[%d]!\n", ret);
        fclose(fp);
        pthread_exit(0);
    }

    for (i = 0; i < requestCnt; i++) {
        while (1) {
            ret = fread(buf, 1, MAX_BUF_SIZE, fp);
            if (ret <= 0)
                break;
            ret = RDT_Write(RDT_ID, buf, ret);
            if (ret < 0) {
                printf("RDT send file data error[%d]!\n", ret);
                pthread_exit(0);
            }

            if ((ret = RDT_Status_Check(RDT_ID, &rdtStatus)) == RDT_ER_NoERROR) {
                if (sInfo.Mode == 2) {
                    if (rdtStatus.BufSizeInSendQueue > 1024000)
                        usleep(50000);
                } else {
                    if (rdtStatus.BufSizeInSendQueue > 512000)
                        usleep(100000);
                    // sleep(1);
                }
            } else {
                printf("RDT status check error[%d]!\n", ret);
                pthread_exit(0);
            }
        }

        fseek(fp, 0, SEEK_SET);
    }

    printf("***RDT_Destroy enter\n");
    RDT_Destroy(RDT_ID);
    printf("thread_Send exit[%d]\n", RDT_ID);
    fclose(fp);

    while (1) {
        struct st_SInfoEx info;
        info.size = sizeof(struct st_SInfoEx);
        if (IOTC_Session_Check_Ex(SID, &info) < 0) {
            IOTC_Session_Close(SID);
            break;
        }
        sleep(1);
    }

    pthread_exit(0);
}

DeviceLoginInput auth_option;
int use_key = 0;

void *_thread_Login(void *arg) {
    int ret;
    char *UID = (char *)arg;
    while (1) {
        if (use_key) {
            ret = IOTC_Device_LoginEx(UID, &auth_option);
            printf("     Calling IOTC_Device_LoginEX() ret = %d\n", ret);
        } else {
            ret = IOTC_Device_Login(UID, NULL,NULL);
            printf("     Calling IOTC_Device_Login() ret = %d\n", ret);
        }

        if (ret == IOTC_ER_NoERROR)
            break;
        else
            sleep(5);
    }

    pthread_exit(0);
}

#define MAX_CLIENT_NUM 8

int main(int argc, char *argv[]) {
    int ret = 0;
    const char *license_key = NULL;

    if (argc < 4) {
        printf("Arg wrong!!!\n");
        printf("  --RDTServer [UID] [Filename] [MaxClientNum]\n");
        printf("  --RDTServer [UID] [Filename] [MaxClientNum] [AuthKey]\n");
        return 0;
    }
    printf("RDT Ver[%s]\n", RDT_GetRDTApiVersionString());

    auth_option.cb = sizeof(DeviceLoginInput);
    auth_option.authentication_type = AUTHENTICATE_BY_KEY;
    memset(auth_option.auth_key, '0', 8);

    if (argc > 4) {
        if(strlen(argv[4]) == 8) {
            strncpy(auth_option.auth_key, argv[4], 8);
            printf("Login with Key=%s\n", argv[4]);
            use_key = 1;
        } else {
            printf("Invalid Key=%s, Login without Key\n", argv[4]);
        }
    } else {
        printf("Login without Key\n");
    }

    int maxClientNum = atoi(argv[3]);
    strcpy(gFn, argv[2]);
    printf("%s\n", gFn);
    gFd = open(gFn, O_RDONLY);
    if (gFd < 0) {
        printf("Open file[%s] error code[%d], %d!!\n", gFn, gFd, errno);
        exit(0);
    }

    //===== set license key for enable TUTK SDK =====
    license_key = getenv(ENV_LICENSE_KEY);
    if (license_key == NULL) {
        printf("Environment variable %s must be set!!\n", ENV_LICENSE_KEY);
        return -1;
    }

    //Set log file path
    LogAttr log_attr;
    memset(&log_attr, 0, sizeof(log_attr));

    log_attr.path = "./log_iotc_device.txt";
    ret = IOTC_Set_Log_Attr(log_attr);
    printf("IOTC_Set_Log_Attr ret[%d]\n", ret);

    log_attr.path = "./log_rdt_device.txt";
    ret = RDT_Set_Log_Attr(log_attr);
    printf("RDT_Set_Log_Attr ret[%d]\n", ret);

    log_attr.path = "./log_global_device.txt";
    ret = TUTK_Set_Log_Attr(log_attr);
    printf("TUTK_Set_Log_Attr ret[%d]\n", ret);

    ret = TUTK_SDK_Set_License_Key(license_key);
    printf("TUTK_SDK_Set_License_Key() ret[%d]\n", ret);
    if (ret != TUTK_ER_NoERROR) {
        return 0;
    }

    ret = IOTC_Initialize2(0);
    if (ret != IOTC_ER_NoERROR) {
        printf("IOTC_Initialize error!!\n");
        return 0;
    }

    int rdtCh = RDT_Initialize();
    if (rdtCh <= 0) {
        printf("RDT_Initialize error!!\n");
        return 0;
    }

    char *UID = (char *)argv[1];
    pthread_t threadID_Login;
    pthread_create(&threadID_Login, NULL, &_thread_Login, (void *)UID);
    pthread_detach(threadID_Login);

    int SID = -1, clientCnt = 0;
    struct st_SInfoEx Sinfo;
    printf("start IOTC_Listen...\n");
    // unsigned long t;
    pthread_t Thread_ID[MAX_CLIENT_NUM];
    do {
        SID = IOTC_Listen(300000);

        if (SID > -1) {
            int *pSID = malloc(sizeof(int));
            if (pSID == NULL) {
                printf("[System error] - malloc failed\n");
                exit(0);
            }
            Sinfo.size = sizeof(struct st_SInfoEx);
            IOTC_Session_Check_Ex(SID, &Sinfo);
            char *mode[3] = {"P2P", "RLY", "LAN"};
            printf("Client from %s:%d Mode=%s\n", Sinfo.RemoteIP, Sinfo.RemotePort, mode[(int)Sinfo.Mode]);
            *pSID = SID;
            pthread_create(&Thread_ID[clientCnt], NULL, &thread_Send, (void *)pSID);
            // if(clientCnt == 0) t = IOTC_GetTickCount();
            clientCnt++;
        }
        if (clientCnt == maxClientNum)
            break;
    } while (1);

    int i;
    for (i = 0; i < maxClientNum; i++)
        pthread_join(Thread_ID[i], NULL);

    // printf("RDT Send Data Cost[%lu ms]....\n", IOTC_GetTickCount()-t);
    RDT_DeInitialize();
    printf("RDT_DeInitialize OK\n");
    IOTC_DeInitialize();

    close(gFd);
    printf("Server exit!\n");

    return 0;
}
